/*
** Libdax
**
** Copyright © 2016 Oracle corp.  All rights reserved.
** The Universal Permissive License (UPL), Version 1.0
**
** Subject to the condition set forth below, permission is hereby granted to any person obtaining a copy of this
** software, associated documentation and/or data (collectively the "Software"), free of charge and under any and
** all copyright rights in the Software, and any and all patent rights owned or freely licensable by each licensor
** hereunder covering either (i) the unmodified Software as contributed to or provided by such licensor, or
** (ii) the Larger Works (as defined below), to deal in both
**
** (a) the Software, and
** (b) any piece of software and/or hardware listed in the lrgrwrks.txt file if one is included with the Software
** (each a “Larger Work” to which the Software is contributed by such licensors),
**
** without restriction, including without limitation the rights to copy, create derivative works of, display,
** perform, and distribute the Software and make, use, sell, offer for sale, import, export, have made, and have
** sold the Software and the Larger Work(s), and to sublicense the foregoing rights on either these or other terms.
**
** This license is subject to the following condition:
** The above copyright notice and either this complete permission notice or at a minimum a reference to the UPL must
** be included in all copies or substantial portions of the Software.
**
** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
** THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
** AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
** CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
** IN THE SOFTWARE.
*/

#ifndef _DAX_H
#define	_DAX_H

/*
 * Header file for user library into which user applications call
 * to take advantage of the DAX (also known as the coprocessor).
 */

/* Currently only the query pipe is supported */

#ifdef	__cplusplus
extern "C" {
#endif

#ifdef __linux
#include <stdint.h>

#ifndef _BOOLEAN_T
#define _BOOLEAN_T
typedef enum { B_FALSE, B_TRUE } boolean_t;
#endif

typedef  unsigned int uint_t;
#define MIN(X,Y) ((X < Y) ? X : Y)
#endif
#include <sys/types.h>
#include <assert.h>
#include <sys/sysmacros.h>
#include <string.h>

/* Available DAX versions to use */
#define	DAX_VERSION_V1 1

/* Initializes a request */
#define	DAX_REQ_INIT(request)					\
	(void) memset(&(request), 0, sizeof (dax_request_t))

/*
 * Convenience macros to help set bit fields correctly.
 * Sample usage: DAX_SETB(hdr, DAX_HDR_SERIAL, 1)
 */

#define	DAX_SETB(x, field, val)     \
	(x = x | (((field##_DATATYPE)(val)) << field##_LOW))

#define	DAX_CLRB(x, field)     \
	x &= (~DAX_MASKB(field))

#define	DAX_MASK(low, high, type)   \
	(((((type)0)-1) >> (sizeof (type) * 8 - (high - low + 1))) << low)

#define	DAX_MASKB(field)    \
	DAX_MASK(field##_LOW, field##_HIGH, field##_DATATYPE)

/* Use DAX_SETB_SLOW if the field has not been previously zero'd */
#define	DAX_SETB_SLOW(x, field, val)	\
	x = ((x) & ~DAX_MASKB(field)) |	\
	    (((field##_DATATYPE)(val)) << field##_LOW)

#define	DAX_GETB(x, field)  \
	(((x) & DAX_MASKB(field)) >> field##_LOW)

/*
 * DAX errors
 *
 * Errors returned by libdax functions.
 */
typedef enum dax_err {
	DAX_SUCCESS = 0,	/* Success */
	DAX_FAILURE,		/* General failure */
	DAX_ENOMEM,		/* No memory available */
	DAX_EWRONGVER,		/* Wrong DAX version */
	DAX_ENOACCESS,		/* Invalid access permissions */
	DAX_ENOTAVAILABLE,	/* DAX not available */
	DAX_ENOTSUPPORTED,	/* DAX not supported */
	DAX_EINITIALIZED,	/* libdax previously initialized */
	DAX_EINVAL,		/* Invalid user arguments */
	DAX_EQTYPE,		/* Invalid queue type */
	DAX_EQFULL,		/* Queue full */
	DAX_EQEMPTY,		/* Queue empty */
	DAX_EBADREQ,		/* Invalid request */
	DAX_ENOREQ,		/* No request for parameters given */
	DAX_EBADALIGN,		/* Address not properly aligned */
	DAX_ESUBMITTED,		/* Request submitted */
	DAX_ENOMAP,		/* No address mapping */
	DAX_EAGAIN,		/* Try again */
	DAX_ENOTSUBMITTED,	/* Request not submitted */
	DAX_ENOTFINISHED	/* Request not finished executing */
} dax_err_t;

/*
 * DAX context
 *
 * Returned when the user initializes libdax.
 */
struct dax_ctx_impl;
typedef struct dax_ctx_impl dax_ctx_t;

/*
 * DAX queue
 *
 * An ordered list of DAX requests that can be submitted to the DAX
 * for execution.
 */
struct dax_queue_impl;
typedef struct dax_queue_impl dax_queue_t;

typedef enum dax_queue_type {
	DAX_QUEUE_TYPE_QUERY
} dax_queue_type_t;

/*
 * DAX machine independent request format: dax_request_t
 */

/*
 * Query flag groups.
 * The flags for dax_query_request_t are grouped into flag-groups,
 * each of size uint64_t. This is done for greater re-usability of macros.
 * New flaggroup_*_t structs may be created for additional flags
 * To reuse a dax_query_request_t instance, zero out all flaggroups
 */

struct dax_query_flaggroup_a {
	uint64_t	sync:8; /* dax_sync_flags_t */

	/* Primary source flags: */
	uint64_t	pri_off:3;	/* primary src bit offset */
	uint64_t	pri_elem_sz:5;	/* primary src element size */
	uint64_t	reserved1:2;
	uint64_t	pri_fmt:4;	/* dax_query_pri_src_fmt_t */
	uint64_t	reserved2:1;
	uint64_t	pri_type_len:2; /* dax_query_pri_src_len_type_t */
	uint64_t	reserved3:1;
	uint64_t	pri_len:24;	/* length of primary source */
	uint64_t	reserved4:14;
};

struct dax_query_flaggroup_b {
	/* Secondary source flags */
	uint64_t	sec_off:3;	/* secondary src offset */
	uint64_t	sec_enc:1;	/* secondary src encoding */
	uint64_t	sec_elem_sz:4;	/* secondary element size (unpacked) */

	/* Select BV flags */
	uint64_t	bv_off:3;	/* Select BV offset */

	/* Output flags */
	uint64_t	out_fmt_out_elem_sz:4;	/* dax_query_out_fmt_t */
	uint64_t	reserved4:2;
	uint64_t	out_buf_blks:20;  /* output buffer 64 byte blocks */
	uint64_t	reserved5:27;
};

struct dax_query_flaggroup_c {
	/* Compression table/Translate	vector flags */
	uint64_t	table_vector_version:2;
	uint64_t	reserved6:4;
	uint64_t	vector_test_value:9;

	/* Scan flags */
	uint64_t	reserved:8;
	uint64_t	scan_op:8;	/* EQ/NE/GE/GT/LT/LE/BETW/NOT_BETW */
	uint64_t	scan_value_1_sz:8;	/* size of value1 */
	uint64_t	scan_value_2_sz:8;	/* size of value2 */

	/* Miscellaneous flags */
	uint64_t	extended_opcode:1;	/* sync/nop command type */
	uint64_t	flow_ctrl:2; /* flow control type */
	uint64_t	reserved7:2; /* pipeline target: NOT USED IN TAHOE */
	uint64_t	pad_dir:1;	/* padding direction */
	uint64_t	reserved8:11;
};

typedef uint64_t	dax_query_flaggroup_a_t;
typedef uint64_t	dax_query_flaggroup_b_t;
typedef uint64_t	dax_query_flaggroup_c_t;

/*
 * In the case when only one dx_scan_value needs to be used,
 * the expectation is that scan_value_1_sz is not 0 while
 * scan_value_2_sz is 0
 */
typedef struct {
	uint8_t			dx_cmd;	/* dax_cmd_t */
	uint8_t			dx_reserved1[7];
	void			*dx_pri_src;	/* pri_src_va */
	void			*dx_sec_src;	/* sec_src_va */
	void			*dx_out;	/* output_va */
	void			*dx_comp_tab; /* comp table, trans vector */
	dax_query_flaggroup_a_t flaggroup_a;
	dax_query_flaggroup_b_t flaggroup_b;
	dax_query_flaggroup_c_t flaggroup_c;
	uint64_t		dx_udata;	/* user data */
	uint64_t		dx_scan_value_1[2]; /* scan value 1 */
	uint64_t		dx_scan_value_2[2]; /* scan value 2 */
	uint64_t		dx_reserved2[3];
} dax_query_request_t;

typedef struct {
	uint8_t		dx_cmd;		/* dax_cmd_t */
	uint8_t		reserved1[7];
	uint64_t	reserved2[7];
	uint64_t	dx_udata;	/* user data */
	uint64_t	reserved3[7];
} dax_any_request_t;

typedef union {
	dax_query_request_t	query;
	dax_any_request_t	any;
} dax_request_t;

/*
 * Bitfield definitions for dax_query_flags_t,
 * used with DAX_SETB() family of macros
 * xx_LOW  - low bit number, origin is 0
 * xx_HIGH - high bit number, inclusive
 * xx_TYPE - type of field (must be unsigned)
 */

/* Query flags */
#define	DAX_QUERY_FLAGS_DATATYPE		uint64_t
#define	DAX_QUERY_FLAGGROUP_A_DATATYPE	DAX_QUERY_FLAGS_DATATYPE
#define	DAX_QUERY_FLAGGROUP_B_DATATYPE	DAX_QUERY_FLAGS_DATATYPE
#define	DAX_QUERY_FLAGGROUP_C_DATATYPE	DAX_QUERY_FLAGS_DATATYPE

/* flaggroup_a definitions */

#define	DAX_QUERY_PRI_SRC_LEN_LOW		14
#define	DAX_QUERY_PRI_SRC_LEN_HIGH		37
#define	DAX_QUERY_PRI_SRC_LEN_DATATYPE		DAX_QUERY_FLAGGROUP_A_DATATYPE

#define	DAX_QUERY_PRI_SRC_LEN_TYPE_LOW		39
#define	DAX_QUERY_PRI_SRC_LEN_TYPE_HIGH		40
#define	DAX_QUERY_PRI_SRC_LEN_TYPE_DATATYPE	DAX_QUERY_FLAGGROUP_A_DATATYPE

#define	DAX_QUERY_PRI_SRC_FMT_LOW		42
#define	DAX_QUERY_PRI_SRC_FMT_HIGH		45
#define	DAX_QUERY_PRI_SRC_FMT_DATATYPE		DAX_QUERY_FLAGGROUP_A_DATATYPE

/* pri_fmt broken down into 4 1-bit fields */
#define	DAX_QUERY_PRI_SRC_FMT_BIT_LOW		42
#define	DAX_QUERY_PRI_SRC_FMT_BIT_HIGH		42
#define	DAX_QUERY_PRI_SRC_FMT_BIT_DATATYPE	DAX_QUERY_FLAGGROUP_A_DATATYPE

#define	DAX_QUERY_PRI_SRC_FMT_VAR_LOW		43
#define	DAX_QUERY_PRI_SRC_FMT_VAR_HIGH		43
#define	DAX_QUERY_PRI_SRC_FMT_VAR_DATATYPE	DAX_QUERY_FLAGGROUP_A_DATATYPE

#define	DAX_QUERY_PRI_SRC_FMT_RLE_LOW		44
#define	DAX_QUERY_PRI_SRC_FMT_RLE_HIGH		44
#define	DAX_QUERY_PRI_SRC_FMT_RLE_DATATYPE	DAX_QUERY_FLAGGROUP_A_DATATYPE

/* SEC_SRC_TYPE is an alias of PRI_SRC_FMT_RLE */
#define	DAX_QUERY_SEC_SRC_TYPE_LOW	DAX_QUERY_PRI_SRC_FMT_RLE_LOW
#define	DAX_QUERY_SEC_SRC_TYPE_HIGH	DAX_QUERY_PRI_SRC_FMT_RLE_HIGH
#define	DAX_QUERY_SEC_SRC_TYPE_DATATYPE	DAX_QUERY_PRI_SRC_FMT_RLE_DATATYPE

#define	DAX_QUERY_PRI_SRC_FMT_HUFF_LOW		45
#define	DAX_QUERY_PRI_SRC_FMT_HUFF_HIGH		45
#define	DAX_QUERY_PRI_SRC_FMT_HUFF_DATATYPE	DAX_QUERY_FLAGGROUP_A_DATATYPE

#define	DAX_QUERY_PRI_SRC_ELEM_SZ_LOW		48
#define	DAX_QUERY_PRI_SRC_ELEM_SZ_HIGH		52
#define	DAX_QUERY_PRI_SRC_ELEM_SZ_DATATYPE	DAX_QUERY_FLAGGROUP_A_DATATYPE

#define	DAX_QUERY_PRI_SRC_OFF_LOW		53
#define	DAX_QUERY_PRI_SRC_OFF_HIGH		55
#define	DAX_QUERY_PRI_SRC_OFF_DATATYPE		DAX_QUERY_FLAGGROUP_A_DATATYPE

#define	DAX_QUERY_SYNC_LOW			56
#define	DAX_QUERY_SYNC_HIGH			63
#define	DAX_QUERY_SYNC_DATATYPE			DAX_QUERY_FLAGGROUP_A_DATATYPE

#define	DAX_QUERY_SERIAL_LOW			56
#define	DAX_QUERY_SERIAL_HIGH			56
#define	DAX_QUERY_SERIAL_DATATYPE		DAX_QUERY_FLAGGROUP_A_DATATYPE

#define	DAX_QUERY_COND_LOW			57
#define	DAX_QUERY_COND_HIGH			57
#define	DAX_QUERY_COND_DATATYPE			DAX_QUERY_FLAGGROUP_A_DATATYPE

#define	DAX_QUERY_LONGCCB_LOW			58
#define	DAX_QUERY_LONGCCB_HIGH			58
#define	DAX_QUERY_LONGCCB_DATATYPE		DAX_QUERY_FLAGGROUP_A_DATATYPE

/* flaggroup_b definitions */

#define	DAX_QUERY_OUT_BUF_BLKS_LOW	27
#define	DAX_QUERY_OUT_BUF_BLKS_HIGH	46
#define	DAX_QUERY_OUT_BUF_BLKS_DATATYPE	DAX_QUERY_FLAGGROUP_B_DATATYPE

#define	DAX_QUERY_OUT_FMT_LOW		49
#define	DAX_QUERY_OUT_FMT_HIGH		52
#define	DAX_QUERY_OUT_FMT_DATATYPE	DAX_QUERY_FLAGGROUP_B_DATATYPE

#define	DAX_QUERY_BV_OFF_LOW		53
#define	DAX_QUERY_BV_OFF_HIGH		55
#define	DAX_QUERY_BV_OFF_DATATYPE	DAX_QUERY_FLAGGROUP_B_DATATYPE

#define	DAX_QUERY_SEC_SRC_ELEM_SZ_LOW	56
#define	DAX_QUERY_SEC_SRC_ELEM_SZ_HIGH	59
#define	DAX_QUERY_SEC_SRC_ELEM_SZ_DATATYPE	DAX_QUERY_FLAGGROUP_B_DATATYPE

#define	DAX_QUERY_SEC_SRC_ENC_LOW	60
#define	DAX_QUERY_SEC_SRC_ENC_HIGH	60
#define	DAX_QUERY_SEC_SRC_ENC_DATATYPE	DAX_QUERY_FLAGGROUP_B_DATATYPE

#define	DAX_QUERY_SEC_SRC_OFF_LOW	61
#define	DAX_QUERY_SEC_SRC_OFF_HIGH	63
#define	DAX_QUERY_SEC_SRC_OFF_DATATYPE	DAX_QUERY_FLAGGROUP_B_DATATYPE

/* flaggroup_c definitions */

#define	DAX_QUERY_PAD_DIR_LOW			11
#define	DAX_QUERY_PAD_DIR_HIGH			11
#define	DAX_QUERY_PAD_DIR_DATATYPE		DAX_QUERY_FLAGGROUP_C_DATATYPE

#define	DAX_QUERY_FLOW_CTRL_LOW			14
#define	DAX_QUERY_FLOW_CTRL_HIGH		15
#define	DAX_QUERY_FLOW_CTRL_DATATYPE		DAX_QUERY_FLAGGROUP_C_DATATYPE

#define	DAX_QUERY_EXTENDED_OPCODE_LOW		16
#define	DAX_QUERY_EXTENDED_OPCODE_HIGH		16
#define	DAX_QUERY_EXTENDED_OPCODE_DATATYPE	DAX_QUERY_FLAGGROUP_C_DATATYPE

#define	DAX_QUERY_SCAN_VALUE_2_SZ_LOW		17
#define	DAX_QUERY_SCAN_VALUE_2_SZ_HIGH		24
#define	DAX_QUERY_SCAN_VALUE_2_SZ_DATATYPE	DAX_QUERY_FLAGGROUP_C_DATATYPE

#define	DAX_QUERY_SCAN_VALUE_1_SZ_LOW		25
#define	DAX_QUERY_SCAN_VALUE_1_SZ_HIGH		32
#define	DAX_QUERY_SCAN_VALUE_1_SZ_DATATYPE	DAX_QUERY_FLAGGROUP_C_DATATYPE

#define	DAX_QUERY_SCAN_OP_LOW			33
#define	DAX_QUERY_SCAN_OP_HIGH			40
#define	DAX_QUERY_SCAN_OP_DATATYPE		DAX_QUERY_FLAGGROUP_C_DATATYPE

#define	DAX_QUERY_VECTOR_TEST_VALUE_LOW		49
#define	DAX_QUERY_VECTOR_TEST_VALUE_HIGH	57
#define	DAX_QUERY_VECTOR_TEST_VALUE_DATATYPE	DAX_QUERY_FLAGGROUP_C_DATATYPE

#define	DAX_QUERY_TABLE_VECTOR_VERSION_LOW	62
#define	DAX_QUERY_TABLE_VECTOR_VERSION_HIGH	63
#define	DAX_QUERY_TABLE_VECTOR_VERSION_DATATYPE	DAX_QUERY_FLAGGROUP_C_DATATYPE

/* Query Pri/Src/Dst VA fields */
#define	DAX_QUERY_ADDR_DATATYPE	uint64_t

#define	DAX_QUERY_PRI_SRC_ADDR_MASK_LOW		55
#define	DAX_QUERY_PRI_SRC_ADDR_MASK_HIGH	59
#define	DAX_QUERY_PRI_SRC_ADDR_MASK_DATATYPE	DAX_QUERY_ADDR_TYPE

#define	DAX_QUERY_SEC_SRC_ADDR_MASK_LOW		55
#define	DAX_QUERY_SEC_SRC_ADDR_MASK_HIGH	59
#define	DAX_QUERY_SEC_SRC_ADDR_MASK_DATATYPE	DAX_QUERY_ADDR_TYPE

#define	DAX_QUERY_OUT_ADDR_MASK_LOW		56
#define	DAX_QUERY_OUT_ADDR_MASK_HIGH		59
#define	DAX_QUERY_OUT_ADDR_MASK_DATATYPE	DAX_QUERY_ADDR_TYPE

#define	DAX_QUERY_COMP_TABLE_ADDR_MASK_LOW	56
#define	DAX_QUERY_COMP_TABLE_ADDR_MASK_HIGH	59
#define	DAX_QUERY_COMP_TABLE_ADDR_MASK_DATATYPE	DAX_QUERY_ADDR_TYPE

/* Query fields */

/* Macros to set parameters */
#define	DAX_REQ_CLRALL_QUERY_FLAGS(req)			\
{							\
	(req)->query.flaggroup_a = 0;			\
	(req)->query.flaggroup_b = 0;			\
	(req)->query.flaggroup_c = 0;			\
}

#define	DAX_REQ_SET_QUERY_CMD(req, val)                     \
	(req)->query.dx_cmd = val

#define	DAX_REQ_SET_QUERY_PRI_SRC(req, val)                 \
	(req)->query.dx_pri_src = val

#define	DAX_REQ_SET_QUERY_SEC_SRC(req, val)                 \
	(req)->query.dx_sec_src = val

#define	DAX_REQ_SET_QUERY_SELECT_BV_SRC(req, val)                 \
	(req)->query.dx_sec_src = val

#define	DAX_REQ_SET_QUERY_OUT(req, val)                 \
	(req)->query.dx_out = val

#define	DAX_REQ_SET_QUERY_COMP_TABLE(req, val)                \
{							\
	(req)->query.dx_comp_tab = val;			\
	DAX_SETB((req)->query.flaggroup_a, DAX_QUERY_PRI_SRC_FMT_HUFF, 1); \
}

#define	DAX_REQ_SET_QUERY_TRANS_VECTOR(req, val)                \
{							\
	(req)->query.dx_comp_tab = val;			\
}
#define	DAX_REQ_SET_QUERY_UDATA(req, val)                   \
	(req)->query.dx_udata = val

#define	DAX_REQ_SET_QUERY_SCAN_VALUE_1(req, hi, lo)	\
	{						\
		(req)->query.dx_scan_value_1[0] = hi;	\
		(req)->query.dx_scan_value_1[1] = lo;	\
	}

#define	DAX_REQ_SET_QUERY_SCAN_VALUE_2(req, hi, lo)	\
	{						\
		(req)->query.dx_scan_value_2[0] = hi;	\
		(req)->query.dx_scan_value_2[1] = lo;	\
	}

#define	DAX_REQ_SET_QUERY_SYNC(req, val)                      \
	DAX_SETB((req)->query.flaggroup_a, DAX_QUERY_SYNC, val)

#define	DAX_REQ_SET_QUERY_PRI_SRC_OFF(req, val)                   \
	DAX_SETB((req)->query.flaggroup_a, DAX_QUERY_PRI_SRC_OFF, val)

#define	DAX_REQ_SET_QUERY_PRI_SRC_ELEM_SZ(req, val)                       \
	DAX_SETB((req)->query.flaggroup_a,			\
	    DAX_QUERY_PRI_SRC_ELEM_SZ, val)

#define	DAX_REQ_SET_QUERY_PRI_SRC_FMT(req, val)                   \
	DAX_SETB((req)->query.flaggroup_a, DAX_QUERY_PRI_SRC_FMT, val)

#define	DAX_REQ_SET_QUERY_PRI_SRC_LEN_TYPE(req, val)                      \
	DAX_SETB((req)->query.flaggroup_a,		\
	    DAX_QUERY_PRI_SRC_LEN_TYPE, val)

#define	DAX_REQ_SET_QUERY_PRI_SRC_LEN(req, val)                   \
	DAX_SETB((req)->query.flaggroup_a, DAX_QUERY_PRI_SRC_LEN, val)

#define	DAX_REQ_SET_QUERY_SEC_SRC_TYPE(req, val)              \
	DAX_SETB((req)->query.flaggroup_a,		\
	    DAX_QUERY_SEC_SRC_TYPE, val)

#define	DAX_REQ_SET_QUERY_SEC_SRC_OFF(req, val)                   \
	DAX_SETB((req)->query.flaggroup_b, DAX_QUERY_SEC_SRC_OFF, val)

#define	DAX_REQ_SET_QUERY_SEC_SRC_ENC(req, val)                   \
	DAX_SETB((req)->query.flaggroup_b, DAX_QUERY_SEC_SRC_ENC, val)

#define	DAX_REQ_SET_QUERY_SEC_SRC_ELEM_SZ(req, val)                       \
	DAX_SETB((req)->query.flaggroup_b,		\
	    DAX_QUERY_SEC_SRC_ELEM_SZ, val)

#define	DAX_REQ_SET_QUERY_SELECT_BV_OFF(req, val)                    \
	DAX_SETB((req)->query.flaggroup_b, DAX_QUERY_BV_OFF, val)

#define	DAX_REQ_SET_QUERY_OUT_FMT(req, val)                   \
	DAX_SETB((req)->query.flaggroup_b, DAX_QUERY_OUT_FMT, val)

#define	DAX_REQ_SET_QUERY_OUT_BUF_BLKS(req, val)                        \
	DAX_SETB((req)->query.flaggroup_b,			\
	    DAX_QUERY_OUT_BUF_BLKS, val)

#define	DAX_REQ_SET_QUERY_TRANS_VECTOR_TEST_VALUE(req, val)                 \
	DAX_SETB((req)->query.flaggroup_c,		\
	    DAX_QUERY_VECTOR_TEST_VALUE, val)

#define	DAX_REQ_SET_QUERY_COMP_TABLE_VERSION(req, val) \
	DAX_SETB((req)->query.flaggroup_c,		\
	    DAX_QUERY_TABLE_VECTOR_VERSION, val)

#define	DAX_REQ_SET_QUERY_TRANS_VECTOR_VERSION(req, val) \
	DAX_SETB((req)->query.flaggroup_c,		\
	    DAX_QUERY_TABLE_VECTOR_VERSION, val)

#define	DAX_REQ_SET_QUERY_SCAN_OP(req, val)                   \
	DAX_SETB((req)->query.flaggroup_c, DAX_QUERY_SCAN_OP, val)

#define	DAX_REQ_SET_QUERY_SCAN_VALUE_1_SZ(req, val)                   \
	DAX_SETB((req)->query.flaggroup_c,		\
	    DAX_QUERY_SCAN_VALUE_1_SZ, val)

#define	DAX_REQ_SET_QUERY_SCAN_VALUE_2_SZ(req, val)                   \
	DAX_SETB((req)->query.flaggroup_c,		\
	    DAX_QUERY_SCAN_VALUE_2_SZ, val)

#define	DAX_REQ_SET_QUERY_EXTENDED_OPCODE(req, val)                  \
	DAX_SETB((req)->query.flaggroup_c,		\
	    DAX_QUERY_EXTENDED_OPCODE, val)

#define	DAX_REQ_SET_QUERY_FLOW_CTRL(req, val)                 \
	DAX_SETB((req)->query.flaggroup_c,		\
	    DAX_QUERY_FLOW_CTRL, val)

#define	DAX_REQ_SET_QUERY_PAD_DIR(req, val)                   \
	DAX_SETB((req)->query.flaggroup_c, DAX_QUERY_PAD_DIR, val)

/* Macros to clear parameters */
#define	DAX_REQ_CLR_QUERY_CMD(req)                  \
	(req)->query.dx_cmd = 0

#define	DAX_REQ_CLR_QUERY_PRI_SRC(req)                      \
	(req)->query.dx_pri_src = 0

#define	DAX_REQ_CLR_QUERY_SEC_SRC(req)                      \
	(req)->query.dx_sec_src = 0

#define	DAX_REQ_CLR_QUERY_SELECT_BV_SRC(req)                      \
	(req)->query.dx_sec_src = 0

#define	DAX_REQ_CLR_QUERY_OUT(req)                      \
	(req)->query.dx_out = 0

#define	DAX_REQ_CLR_QUERY_COMP_TABLE(req)             \
	(req)->query.dx_comp_tab = 0

#define	DAX_REQ_CLR_QUERY_TRANS_VECTOR(req)             \
	(req)->query.dx_comp_tab = 0

#define	DAX_REQ_CLR_QUERY_UDATA(req)                        \
	(req)->query.dx_udata = 0

#define	DAX_REQ_CLR_QUERY_SCAN_VALUE_1(req)		\
	{						\
		(req)->query.dx_scan_value_1[0] = 0;	\
		(req)->query.dx_scan_value_1[1] = 0;	\
	}

#define	DAX_REQ_CLR_QUERY_SCAN_VALUE_2(req)		\
	{						\
		(req)->query.dx_scan_value_2[0] = 0;	\
		(req)->query.dx_scan_value_2[1] = 0;	\
	}

#define	DAX_REQ_CLR_QUERY_SYNC(req)                   \
	DAX_CLRB((req)->query.flaggroup_a, DAX_QUERY_SYNC)

#define	DAX_REQ_CLR_QUERY_PRI_SRC_OFF(req)                        \
	DAX_CLRB((req)->query.flaggroup_a, DAX_QUERY_PRI_SRC_OFF)

#define	DAX_REQ_CLR_QUERY_PRI_SRC_ELEM_SZ(req)                    \
	DAX_CLRB((req)->query.flaggroup_a,		\
	    DAX_QUERY_PRI_SRC_ELEM_SZ)

#define	DAX_REQ_CLR_QUERY_PRI_SRC_FMT(req)                        \
	DAX_CLRB((req)->query.flaggroup_a, DAX_QUERY_PRI_SRC_FMT)

#define	DAX_REQ_CLR_QUERY_PRI_SRC_LEN_TYPE(req)                   \
	DAX_CLRB((req)->query.flaggroup_a,		\
	    DAX_QUERY_PRI_SRC_LEN_TYPE)

#define	DAX_REQ_CLR_QUERY_PRI_SRC_LEN(req)                        \
	DAX_CLRB((req)->query.flaggroup_a, DAX_QUERY_PRI_SRC_LEN)

#define	DAX_REQ_CLR_QUERY_SEC_SRC_TYPE(req)                   \
	DAX_CLRB((req)->query.flaggroup_a,		\
	    DAX_QUERY_SEC_SRC_TYPE)

#define	DAX_REQ_CLR_QUERY_SEC_SRC_OFF(req)                        \
	DAX_CLRB((req)->query.flaggroup_b, DAX_QUERY_SEC_SRC_OFF)

#define	DAX_REQ_CLR_QUERY_SEC_SRC_ENC(req)                        \
	DAX_CLRB((req)->query.flaggroup_b, DAX_QUERY_SEC_SRC_ENC)

#define	DAX_REQ_CLR_QUERY_SEC_SRC_ELEM_SZ(req)                    \
	DAX_CLRB((req)->query.flaggroup_b,		\
	    DAX_QUERY_SEC_SRC_ELEM_SZ)

#define	DAX_REQ_CLR_QUERY_SELECT_BV_OFF(req)                 \
	DAX_CLRB((req)->query.flaggroup_b, DAX_QUERY_BV_OFF)

#define	DAX_REQ_CLR_QUERY_OUT_FMT(req)                        \
	DAX_CLRB((req)->query.flaggroup_b,		\
	    DAX_QUERY_OUT_FMT)

#define	DAX_REQ_CLR_QUERY_OUT_BUF_BLKS(req)                     \
	DAX_CLRB((req)->query.flaggroup_b,		\
	    DAX_QUERY_OUT_BUF_BLKS)

#define	DAX_REQ_CLR_QUERY_TRANS_VECTOR_TEST_VALUE(req)                      \
	DAX_CLRB((req)->query.flaggroup_c,		\
	    DAX_QUERY_VECTOR_TEST_VALUE)

#define	DAX_REQ_CLR_QUERY_COMP_TABLE_VERSION(req)                   \
	DAX_CLRB((req)->query.flaggroup_c,		\
	    DAX_QUERY_TABLE_VECTOR_VERSION)

#define	DAX_REQ_CLR_QUERY_TRANS_VECTOR_VERSION(req)                   \
	DAX_CLRB((req)->query.flaggroup_c,		\
	    DAX_QUERY_TABLE_VECTOR_VERSION)

#define	DAX_REQ_CLR_QUERY_SCAN_OP(req)                        \
	DAX_CLRB((req)->query.flaggroup_c, DAX_QUERY_SCAN_OP)

#define	DAX_REQ_CLR_QUERY_SCAN_VALUE_1_SZ(req)                        \
	DAX_CLRB((req)->query.flaggroup_c,		\
	    DAX_QUERY_SCAN_VALUE_1_SZ)

#define	DAX_REQ_CLR_QUERY_SCAN_VALUE_2_SZ(req)                        \
	DAX_CLRB((req)->query.flaggroup_c,		\
	    DAX_QUERY_SCAN_VALUE_2_SZ)

#define	DAX_REQ_CLR_QUERY_EXTENDED_OPCODE(req)                       \
	DAX_CLRB((req)->query.flaggroup_c, DAX_QUERY_EXTENDED_OPCODE)

#define	DAX_REQ_CLR_QUERY_FLOW_CTRL(req)                      \
	DAX_CLRB((req)->query.flaggroup_c,		\
	    DAX_QUERY_FLOW_CTRL)

#define	DAX_REQ_CLR_QUERY_PAD_DIR(req)                        \
	DAX_CLRB((req)->query.flaggroup_c, DAX_QUERY_PAD_DIR)

/*
 * DAX machine independent result area format: dax_result_t
 */
typedef struct {
	uint8_t	dxr_status;	/* Request Status */
	uint8_t	dxr_errcode;	/* Request error */
	uint8_t dxr_submit_error; /* Submit error, dax_err_t */
	uint8_t	dxr_reserved1;
	uint32_t dxr_err_value;	/* Request error value */
	uint64_t dxr_cycles;	/* Runtime */
	uint64_t dxr_udata;	/* User data */
	uint64_t dxr_retval;	/* Command return value */
	uint32_t dxr_ilen;	/* input elements consumed */
	uint32_t dxr_olen;	/* output size in bytes */
	uint64_t dxr_reserved2[3];
} dax_query_result_t;

typedef struct {
	uint8_t	dxr_status;
	uint8_t	dxr_errcode;
	uint8_t dxr_submit_error; /* dax_err_t */
	uint8_t	dxr_reserved1[5];
	uint64_t dxr_cycles;
	uint64_t dxr_udata;
	uint64_t dxr_retval;
	uint64_t dxr_reserved2[4];
} dax_any_result_t;

typedef union {
	dax_query_result_t 	query;
	dax_any_result_t	any;
} dax_result_t;

/*
 * Macros to return result fields
 */
#define	DAX_RES_GET_QUERY_SUBMIT_ERR(res)	\
	(dax_err_t)((res)->query.dxr_submit_error)
#define	DAX_RES_GET_QUERY_STATUS(res)	\
	(uint8_t)((res)->query.dxr_status)
#define	DAX_RES_GET_QUERY_ERRCODE(res)	\
	(uint8_t)((res)->query.dxr_errcode)
#define	DAX_RES_GET_QUERY_ERR_VALUE(res)	\
	(uint32_t)((res)->query.dxr_err_value);
#define	DAX_RES_GET_QUERY_RUNTIME(res)	\
	(uint64_t)((res)->query.dxr_cycles)
#define	DAX_RES_GET_QUERY_UDATA(res)	\
	(uint64_t)((res)->query.dxr_udata)
#define	DAX_RES_GET_QUERY_RETVAL(res)	\
	(uint64_t)((res)->query.dxr_retval)
#define	DAX_RES_GET_QUERY_NPROCESSED(res)	\
	(uint64_t)((res)->query.dxr_ilen)
#define	DAX_RES_GET_QUERY_OUT_LEN(res)	\
	(uint64_t)((res)->query.dxr_olen)

/*
 *
 * Request enumerations
 *
 */

/*
 * Request enumerations for query requests
 */

/* Commands for query requests */
typedef enum dax_cmd {
	DAX_CMD_QUERY_SYNC_NOP = 0x00,
	DAX_CMD_QUERY_EXTRACT = 0x01,
	DAX_CMD_QUERY_SCAN = 0x02,
	DAX_CMD_QUERY_TRANSLATE = 0x04,
	DAX_CMD_QUERY_INV_TRANSLATE = 0x14,
	DAX_CMD_QUERY_SELECT = 0x05
} dax_cmd_t;

/* Sync flags */
typedef enum dax_sync_flags {
	DAX_SYNC_FLAGS_OFF = 0x00,
	DAX_SYNC_FLAGS_SERIAL = 0x01,
	DAX_SYNC_FLAGS_CONDITIONAL = 0x02,
	DAX_SYNC_FLAGS_SERIAL_CONDITIONAL = 0x03
} dax_sync_flags_t;

/* Sync nop command type */
typedef enum dax_sync_nop_cmd_type {
	DAX_SYNC_NOP_CMD_TYPE_NOP = 0x00,
	DAX_SYNC_NOP_CMD_TYPE_SYNC = 0x01
} dax_sync_nop_cmd_type_t;


/*
 * Query request enumerations
 */

/* Query primary source format */
typedef enum dax_query_pri_src_fmt {
	DAX_QUERY_PRI_SRC_FMT_FIXED_WIDTH_BYTE = 0x00,
	DAX_QUERY_PRI_SRC_FMT_FIXED_WIDTH_BIT = 0x01,
	DAX_QUERY_PRI_SRC_FMT_VAR_WIDTH_BYTE = 0x02
} dax_query_pri_src_fmt_t;

/* Query primary source length type */
typedef enum dax_query_pri_src_len_type {
	DAX_QUERY_PRI_SRC_LEN_TYPE_SYMBOLS = 0x00,
	DAX_QUERY_PRI_SRC_LEN_TYPE_BYTES = 0x01,
	DAX_QUERY_PRI_SRC_LEN_TYPE_BITS = 0x02
} dax_query_pri_src_len_type_t;

/* Query secondary source type */
typedef enum dax_query_sec_src_type {
	DAX_QUERY_SEC_SRC_TYPE_LEN_STREAM = 0x00,
	DAX_QUERY_SEC_SRC_TYPE_RLE_STREAM = 0x01
} dax_query_sec_src_type_t;

/* Query secondary source encoding */
typedef enum dax_query_sec_src_enc {
	DAX_QUERY_SEC_SRC_ENC_ACTUAL_LESS_ONE = 0x00,
	DAX_QUERY_SEC_SRC_ENC_ACTUAL = 0x01
} dax_query_sec_src_enc_t;

/* Query output format */
typedef enum dax_query_output_fmt {
	DAX_QUERY_OUTPUT_FMT_ALIGNED_1BYTE = 0x00,
	DAX_QUERY_OUTPUT_FMT_ALIGNED_2BYTE = 0x01,
	DAX_QUERY_OUTPUT_FMT_ALIGNED_4BYTE = 0x02,
	DAX_QUERY_OUTPUT_FMT_ALIGNED_8BYTE = 0x03,
	DAX_QUERY_OUTPUT_FMT_ALIGNED_16BYTE = 0x04,
	DAX_QUERY_OUTPUT_FMT_BIT_VEC = 0x08,
	DAX_QUERY_OUTPUT_FMT_ONE_INDEX_2BYTE = 0x0d,
	DAX_QUERY_OUTPUT_FMT_ONE_INDEX_4BYTE = 0x0e
} dax_query_output_fmt_t;

/* Query flow control type */
typedef enum dax_query_flow_ctl_type {
	DAX_QUERY_FLOW_CTL_TYPE_DISABLE = 0x00,
	DAX_QUERY_FLOW_CTL_TYPE_BUF_LIMIT = 0x02
} dax_query_flow_ctl_type_t;

/* Query pad direction */
typedef enum dax_query_pad_dir {
	DAX_QUERY_PAD_DIR_RIGHT = 0x00,
	DAX_QUERY_PAD_DIR_LEFT = 0x01
} dax_query_pad_dir_t;

/*
 * Query scan value ops are defined as follows (where a is the
 * bound and x is the value):
 * EQ: x == a
 * NE: x != a
 *
 * Query scan bound ops are defined as follows (where a is the
 * bound and x is the value):
 *
 * GT: x > a
 * GE: x >= a
 * LT: x < a
 * LE: x <= a
 *
 * Query scan range types are defined as follows (where b1 is bound 1,
 * b2 is bound 2, and x is the value):
 *
 * BETWEEN: b1 <= x <= b2
 * NOT_BETWEEN: x < b1 || x > b2
 */
typedef enum dax_query_scan_op {
	DAX_QUERY_SCAN_OP_EQ,
	DAX_QUERY_SCAN_OP_NE,
	DAX_QUERY_SCAN_OP_GT,
	DAX_QUERY_SCAN_OP_GE,
	DAX_QUERY_SCAN_OP_LT,
	DAX_QUERY_SCAN_OP_LE,
	DAX_QUERY_SCAN_OP_BETWEEN,
	DAX_QUERY_SCAN_OP_NOT_BETWEEN
} dax_query_scan_op_t;

/*
 *
 * Result enumerations
 *
 */

/*
 * Query result enumerations
 */

/* Query request errors (returned by hardware) */
typedef enum dax_query_error {
	DAX_QUERY_ERROR_FATAL = 0x0,
	DAX_QUERY_ERROR_BUFFER_OVERFLOW = 0x01,
	DAX_QUERY_ERROR_CCB_DECODE = 0x02,
	DAX_QUERY_ERROR_PAGE_OVERFLOW = 0x03,
	DAX_QUERY_ERROR_CMD_KILLED = 0x07,
	DAX_QUERY_ERROR_TIMEOUT = 0x08,
	DAX_QUERY_ERROR_ADI = 0x09,
	DAX_QUERY_ERROR_DATA_FORMAT = 0x0a,
	DAX_QUERY_ERROR_NO_RETRY = 0x0e,
	DAX_QUERY_ERROR_RETRY = 0x0f,
	DAX_QUERY_ERROR_PARTIAL_SYMBOL_WARNING = 0x80
} dax_query_error_t;

/*
 * DAX status
 *
 * Hardware execution status of a request.
 */
typedef enum dax_status {
	DAX_STATUS_INPROGRESS = 0,
	DAX_STATUS_SUCCESS = 1,
	DAX_STATUS_ERROR = 2,
	DAX_STATUS_KILLED = 3,
	DAX_STATUS_NOTRUN = 4
} dax_status_t;

/* DAX log levels */
typedef enum dax_log {
	DAX_LOG_NONE = 0,
	DAX_LOG_FATAL,
	DAX_LOG_ERROR,
	DAX_LOG_WARN,
	DAX_LOG_INFO,
	/* The DEBUG level logs are printed only in the debug build of libdax */
	DAX_LOG_DEBUG
} dax_log_t;

/* Checks if the DAX can be used */
boolean_t dax_check(void);

/* Initializes libdax for a thread and creates a context for further use */
dax_ctx_t *dax_thread_init(uint32_t version, dax_err_t *err);

/* Gets the file descriptor associated with a context */
int dax_get_fd(dax_ctx_t *ctx, dax_err_t *err);

/* Creates a queue of a given type */
dax_queue_t *dax_create_queue(dax_ctx_t *ctx, dax_queue_type_t queue_type,
    size_t *queue_size, dax_err_t *err);

/* Queues a request */
dax_err_t dax_queue_request(dax_queue_t *queue, dax_request_t *request);

/* Cancels a request before submission for execution */
dax_err_t dax_cancel_request(dax_queue_t *queue, uint64_t udata);

/* Submits all queued requests for execution */
uint64_t dax_execute(dax_queue_t *queue, dax_err_t *err);

/*
 * Reads the results of first num_results requests on the queue
 * within the given timeout.
 */
uint64_t dax_read_results(dax_queue_t *queue, dax_result_t *results,
    uint64_t num_results, uint64_t timeout, dax_err_t *err);

/* Destroys a queue */
dax_err_t dax_destroy_queue(dax_queue_t *queue);

/* Finishes use of libdax */
dax_err_t dax_thread_fini(dax_ctx_t *ctx);

/* Returns error string for given dax_err_t */
char *dax_strerror(dax_err_t err);

/* Sets a logging level */
dax_err_t dax_log_set_level(dax_log_t log_level);

/* formats a page size */
char *physpage_str(void *addr);

/* allocate contiguous ram for use as dax output */
void *dax_alloc_ram(dax_ctx_t *ctx, size_t length, dax_err_t *err);

/* allocate contiguous ram for use as dax output (callable from libdax-query) */
void *dax_alloc_segment(dax_queue_t *queue, size_t length, dax_err_t *err);

/* deallocate contiguous ram */
void dax_free_ram(dax_ctx_t *ctx, void *va, size_t length, dax_err_t *err);

/* deallocate contiguous ram (callable from libdax_query */
void dax_free_segment(dax_queue_t *queue, void *va, size_t length, dax_err_t *err);

#ifdef	__cplusplus
}
#endif

#endif	/* _DAX_H */
